import ee
import pandas as pd
import plotly.graph_objects as go

from sankee.datasets import Dataset
from sankee import utils

# Temporary property name to store image labels in
LABEL_PROPERTY = "sankee_label"


def _label_images(image_list, label_list):
    """
    Take a list of images and assign provided or auto-generated labels to each. Return the labeled images and the
    list of labels.
    """
    if not label_list:
        label_list = [i for i in range(len(image_list))]

    # Assign a label to one image in a list of images
    def apply_label(img, img_list):
        img_list = ee.List(img_list)

        index = img_list.indexOf(img)
        img_label = ee.List(label_list).get(index)

        indexed_img = ee.Image(img).set(LABEL_PROPERTY, img_label)
        img_list = img_list.set(img_list.indexOf(img), indexed_img)

        return img_list

    labeled_images = ee.ImageCollection(
        ee.List(ee.List(image_list).iterate(apply_label, image_list)))

    return labeled_images, label_list


def _extract_values(image_list, samples, band, scale):
    """
    Take a list of images and a collection of sample points and extract image values to each sample point. The image
    values will be stored in a property based on the image label.
    """
    def extract_values(point):
        # Set the location to extract values at
        geom = point.geometry()

        def extract(img, feature):
            # Extract the cover value at the point
            cover = img.reduceRegion(ee.Reducer.first(), geom, scale).get(band)
            # Get the user-defined label that was stored in the image
            label = img.get(LABEL_PROPERTY)

            # Set a property where the name is the label and the value is the extracted cover
            return ee.Feature(feature).set(label, cover)

        return ee.Feature(image_list.iterate(extract, point))

    sample_data = samples.map(extract_values)

    return sample_data


def _sample(image_list, region, dataset, label_list=None, n=100, scale=None, seed=0):
    """
    Randomly sample values of a list of images to quantify change over time.

    :param list image_list: An list of classified ee.Image objects representing change over time.
    :param ee.Geometry region: The region to sample.
    :param geevis.datasets.Dataset dataset: A dataset to which the start and end images belong, which contains a band
    value. If a dataset is not provided, a band name must be explicity provided.
    :param str band: The name of the band of start_img and end_img that contains the class value. Not needed if a
    dataset parameter is provided.
    :param list label_list: A list of labels associated with each image in image_list. If not provided, numeric labels
    will be automatically generated starting at 0.
    :param int n: The number of sample points to extract. More points will take longer to run but generate more
    representative cover statistics.
    :param int scale: The scale to sample point statistics at. Generally, this should be the nominal scale of the
    start and end image.
    :param int seed: Random seed used to generate sample points.
    :return pd.DataFrame: A dataframe in which each row represents as single sample point and columns represent the
    class of that point in each image of the image list.
    """
    # Apply labels to images
    labeled_images, label_list = _label_images(image_list, label_list)

    # Create sample points
    samples = ee.FeatureCollection.randomPoints(region, n, seed)

    # Extract image values at each sample point
    sample_data = _extract_values(labeled_images, samples, dataset.band, scale)

    try:
        data = pd.DataFrame.from_dict(
            [feat["properties"] for feat in ee.Feature(sample_data).getInfo()["features"]])
    except ee.EEException as e:
        # ee may raise an error if the band name isn't valid. This is a bit of a hack to catch that since ee doesn't
        # raise more specific errors.
        if dataset.band in e.args[0]:
            raise ValueError(
                f'"{dataset.band}" is not a valid band name. Check that the dataset band name exists for all images.')
        else:
            raise e

    return data[label_list]


def _reformat(data, dataset):
    """
    Take a dataframe of data representing classified sample points and return all parameters needed to generate a
    Sankey plot. This is done by looping through columns in groups of two representing start and end conditions and
    reformating data to match the Plotly Sankey input parameters.

    :param pd.DataFrame data: A dataframe in which each row represents as single sample point and columns represent
    classes over an arbitrary number of time periods.
    :param geevis.dataset.Dataset dataset: A dataset from which the class data was generated, containing labels and
    palettes corresponding to class values. If a dataset is not provided, class labels and a class palette must be
    provided instead.
    :param dict labels: A dictionary where keys are the class index values and the values are corresponding
    labels. Every class index in the sample dataset must be included in labels.
    :param dict palette: A dictionary where keys are the class index values and the values are corresponding
    colors. Every class index in the sample dataset must be included in palette.
    :return tuple: A tuple of values used in Sankey plotting in the following order: node labels, link labels, node
    palette, link palette, labels, source, target, and values.
    """
    # Take a dataframe with two columns representing start conditions and end conditions and return it in a new dataframe
    # for use in sankey plotting
    def reformat_group(group_data, start_index=0):
        column_list = group_data.columns.tolist()

        # Transform the data to get counts of each combination of condition
        sankey_data = group_data.groupby(
            column_list).size().reset_index(name="value")

        # Calculate normalized change from start to end condition
        sankey_data["change"] = utils.normalized_change(
            sankey_data, column_list[0], "value")

        # Get lists of unique source and target classes
        unique_source = pd.unique(
            data[column_list[0]].values.flatten()).tolist()
        unique_target = pd.unique(
            data[column_list[1]].values.flatten()).tolist()

        # Generate a unique index for each source and target
        sankey_data["source"] = sankey_data[column_list[0]].apply(
            lambda x: unique_source.index(x) + start_index)
        # Offset the target IDs by the last source class to prevent overlap with source IDs
        sankey_data["target"] = sankey_data[column_list[1]].apply(
            lambda x: unique_target.index(x) + sankey_data.source.max() + 1)

        # Assign labels to each source and target
        sankey_data["source_label"] = sankey_data[column_list[0]].apply(
            lambda i: dataset.labels[i])
        sankey_data["target_label"] = sankey_data[column_list[1]].apply(
            lambda i: dataset.labels[i])

        return sankey_data[["source", "target", "value", "source_label", "target_label", "change"]]

    # Calculate the max index that will be needed based on the number of unique classes in all groups
    max_id = sum([len(pd.unique(col)) for _, col in data.iteritems()])

    # Pre-allocate a list of labels that will be iteratively assigned labels up to the max id
    label = [None for i in range(max_id)]

    node_labels = []
    link_labels = []
    source = []
    target = []
    value = []
    current_index = 0

    for i in range(data.columns.size - 1):
        column_group = (range(i, i+2))
        # Select a set of start and end condition columns
        group_data = data.iloc[:, column_group]

        sankified = reformat_group(group_data, start_index=current_index)
        # The start index of the next column group will be the end index of this column group. This sets the index
        # offset to achieve that.
        current_index = sankified.target.min()

        start_label = group_data.columns[0]
        end_label = group_data.columns[1]

        # Store the column label for the source data
        node_labels += [start_label for i in range(
            len(pd.unique(sankified.source)))]

        # Generate a list of strings describing the change in each row
        for i, row in sankified.iterrows():
            if row.source_label == row.target_label:
                link_label = f"{round(row.change * 100)}% of {row.source_label} remained {row.target_label}"
            else:
                link_label = f"{round(row.change * 100)}% of {row.source_label} became {row.target_label}"
            link_labels.append(link_label)

            # Assign class labels to match their respective indexes
            label[row.source] = row.source_label
            label[row.target] = row.target_label

        source += sankified.source.tolist()
        target += sankified.target.tolist()
        value += sankified.value.tolist()

    # Store the column label for the final target data
    node_labels += [end_label for i in range(len(pd.unique(sankified.target)))]
    node_palette = [dataset.get_color(l) for l in label]
    link_palette = [dataset.get_color(i) for i in [label[j] for j in source]]

    return (node_labels, link_labels, node_palette, link_palette, label, source, target, value)


def _clean(data, exclude=None, max_classes=None, dropna=True):
    """
    Perform some cleaning on data before plotting by excluding unwanted classes and limiting the number of classes.

    :param pd.DataFrame data: A dataframe in which each row represents as single sample point and columns represent the
    class of that point in each image of an image list.
    :param list exclude: A list of class values to remove from the dataframe.
    :param int max_classes: The maximum number of unique classes to include in the dataframe. If more classes are present,
    the smallest classes will be omitted from the plot. If max_classes is None, no classes will be dropped.
    :param bool dropna: If true, samples with no class data in any column will be dropped.
    :return pd.DataFrame: The input dataframe with cleaning applied.
    """
    if dropna:
        data = data.dropna()
    if exclude:
        data = data[~data.isin(exclude).any(axis=1)]
    if max_classes:
        data = utils.drop_classes(data, max_classes)

    return data


def _plot(node_labels, link_labels, node_palette, link_palette, label, source, target, value, title=None):
    """
    Generate a Sankey plot of land cover change over an arbitrary number of time steps.
    """
    fig = go.Figure(data=[go.Sankey(
        node=dict(
            pad=30,
            thickness=20,
            line=dict(color="#000000", width=1),
            customdata=node_labels,
            hovertemplate='%{customdata}<extra></extra>',
            label=label,
            color=node_palette
        ),
        link=dict(
            source=source,
            target=target,
            line=dict(color="#909090", width=1),
            value=value,
            color=link_palette,
            customdata=link_labels,
            hovertemplate='%{customdata} <extra></extra>',
        ))])

    fig.update_layout(
        title_text=f"<b>{title}</b>" if title else None,
        font_size=16,
        title_x=0.5,
        paper_bgcolor='rgba(0, 0, 0, 0)'
    )

    return fig


def sankify(image_list, region, label_list=None, dataset=None, band=None, labels=None, palette=None,
            exclude=None, max_classes=None, n=100, title=None, scale=None, seed=0, dropna=True):
    """
    Perform sampling, data cleaning and reformatting, and generation of a Sankey plot of land cover change over time
    within a region.
    """
    dataset = utils.build_dataset(dataset, band, labels, palette)
    label_list = utils.build_label_list(image_list, label_list)

    utils.test_params(dataset, image_list, label_list)

    data = _sample(image_list, region, dataset=dataset, label_list=label_list,
                   n=n, scale=scale, seed=seed)

    utils.check_plot_params(data, dataset)
    cleaned = _clean(data, exclude, max_classes, dropna=dropna)
    node_labels, link_labels, node_palette, link_palette, label, source, target, value = _reformat(
        cleaned, dataset)
    return _plot(node_labels, link_labels, node_palette, link_palette, label, source, target, value, title=title)



import numpy as np
from sankee.datasets import Dataset


def get_missing_keys(key_list, key_dict):
    """
    Find any keys that are present in a list that are not present in the keys of a dictionary. Helpful for testing if a 
    label or palette dictionary is fully defined.
    """
    missing = []
    for key in key_list:
        if key not in key_dict.keys():
            missing.append(key)
    return missing


def normalized_change(data, group_col, count_col):
    """
    Perform group-wise data normalization to a column of values. Output values will be the proportion of all values 
    in the group.
    """
    return data.groupby(group_col)[count_col].apply(lambda x: x / x.sum())


def check_plot_params(data, dataset):
    """
    Check for values that are present in data and are not present in labels or palette and raise an error if any are
    found.
    """
    missing_labels = []
    missing_palette = []

    for _, col in data.iteritems():
        missing_labels += get_missing_keys(col, dataset.labels)
        missing_palette += get_missing_keys(col, dataset.palette)

    if missing_labels:
        raise Exception(
            f"The following values are present in the data and undefined in the labels: {np.unique(missing_labels)}")
    if missing_palette:
        raise Exception(
            f"The following values are present in the data and undefined in the palette: {np.unique(missing_palette)}")


def test_params(dataset, image_list, label_list):
    """
    Run tests to ensure user parameters are usable.
    """
    if not dataset.band:
        raise ValueError("Provide dataset or band.")
    if not dataset.labels:
        raise ValueError("Provide dataset or labels.")
    if not dataset.palette:
        raise ValueError("Provide dataset or palette.")
    if len(label_list) != len(image_list):
        raise ValueError(
            "Length of label list must match length of image list.")
    return 0


def drop_classes(data, max_classes, metric="area"):
    """
    Remove small classes until a maximum number of classes is reached.

    :param pd.DataFrame data: A dataframe in which each row represents as single sample point and columns represent
    the class of that point at various times.
    :param int max_classes: The maximum number of unique classes to retain. If more classes are present, the smallest
    classes will be removed.
    :return pd.DataFrame: A dataframe with rows that belong to the largest classes.
    """
    class_counts = data.melt().groupby("value").size().reset_index(name="n")
    largest_classes = class_counts.sort_values(
        by="n", ascending=False).value[:max_classes].tolist()
    dropped_data = data[data.isin(largest_classes).all(axis=1)]

    return dropped_data


def build_dataset(dataset=None, band="", labels=None, palette=None):
    """
    Take a dataset and/or some combination of band, labels, and palette and return a dataset. If a dataset is
    provided, a copy will be returned with any provided parameters replaced.
    """
    labels = labels if labels else {}
    palette = palette if palette else {}

    # Replace any dataset parameters with provided parameters
    if dataset:
        band = band if band else dataset.band
        labels = labels if labels else dataset.labels
        palette = palette if palette else dataset.palette

    built = Dataset(collection_name=None, band=band,
                    labels=labels, palette=palette)

    return built


def build_label_list(image_list, label_list=None):
    """
    Take an image list and an optional label list. If a label list is provided, elements will be cast to string. If not,
    a label list will be generated using sequential numbers to match the image list.
    """
    if not label_list:
        label_list = [i for i in range(len(image_list))]
    # Cast every element in label list to string since GEE can't handle that
    label_list = [str(x) for x in label_list]

    return label_list



from enum import Enum
import ee
import pandas as pd


class Dataset:
    def __init__(self, collection_name, band, labels, palette):
        """
        :param str band: The name of the image band that contains class values.
        :param dict labels: A dictionary matching class values to their corresponding labels.
        :param dict palette: A dictioanry matching class values to their corresponding hex colors.
        """
        self.collection_name = collection_name
        self.band = band
        self.labels = labels
        self.palette = palette

    def __repr__(self):
        return f"<sankee.datasets.Dataset> {self.title}" \


    @property
    def keys(self):
        return list(self.labels.keys())

    @property
    def df(self):
        """
        Return a Pandas dataframe describing the dataset parameters
        """
        return pd.DataFrame({"id": self.keys, "label": self.labels.values(), "color": self.palette.values()})

    @property
    def title(self):
        return self.collection.get("title").getInfo()

    @property
    def collection(self):
        return ee.ImageCollection(self.collection_name)

    def id(self):
        return self.collection.get("system:id").getInfo()

    def get_color(self, label):
        """
        Take a label and return the associated color from the palette.
        """
        label_key = [k for k, v in self.labels.items() if v == label][0]
        return self.palette[label_key]

    def get_images(self, max_images=20):
        """
        List the names of the first n images in the dataset collection up to max_images
        """
        img_list = []
        for img in self.collection.toList(max_images).getInfo():
            try:
                img_list.append(img["id"])
            except KeyError:
                pass

        if len(img_list) == max_images:
            img_list.append('...')
        return img_list


class datasets(Dataset, Enum):
    NLCD2016 = (
        "USGS/NLCD",
        "landcover", {
            1: "No data",
            11: "Open water",
            12: "Perennial ice/snow",
            21: "Developed, open space",
            22: "Developed, low intensity",
            23: "Developed, medium intensity",
            24: "Developed, high intensity",
            31: "Barren land (rock/sand/clay)",
            41: "Deciduous forest",
            42: "Evergreen forest",
            43: "Mixed forest",
            51: "Dwarf scrub",
            52: "Shrub/scrub",
            71: "Grassland/herbaceous",
            72: "Sedge/herbaceous",
            73: "Lichens",
            74: "Moss",
            81: "Pasture/hay",
            82: "Cultivated crops",
            90: "Woody wetlands",
            95: "Emergent herbaceous wetlands"
        }, {
            1: "#000000",
            11: "#466b9f",
            12: "#d1def8",
            21: "#dec5c5",
            22: "#d99282",
            23: "#eb0000",
            24: "#ab0000",
            31: "#b3ac9f",
            41: "#68ab5f",
            42: "#1c5f2c",
            43: "#b5c58f",
            51: "#af963c",
            52: "#ccb879",
            71: "#dfdfc2",
            72: "#d1d182",
            73: "#a3cc51",
            74: "#82ba9e",
            81: "#dcd939",
            82: "#ab6c28",
            90: "#b8d9eb",
            95: "#6c9fb8"
        })

    MODIS_LC_TYPE1 = (
        "MODIS/006/MCD12Q1",
        "LC_Type1",
        {
            1: "Evergreen conifer forest",
            2: "Evergreen broadleaf forest",
            3: "Deciduous conifer forest",
            4: "Deciduous broadleaf forest",
            5: "Mixed forest",
            6: "Closed shrubland",
            7: "Open shrubland",
            8: "Woody savanna",
            9: "Savanna",
            10: "Grassland",
            11: "Permanent wetland",
            12: "Cropland",
            13: "Urban",
            14: "Cropland and natural vegetation",
            15: "Permanent snow and ice",
            16: "Barren",
            17: "Water"
        },
        {
            1: "#086a10",
            2: "#dcd159",
            3: "#54a708",
            4: "#78d203",
            5: "#009900",
            6: "#c6b044",
            7: "#dcd159",
            8: "#dade48",
            9: "#fbff13",
            10: "#b6ff05",
            11: "#27ff87",
            12: "#c24f44",
            13: "#a5a5a5",
            14: "#ff6d4c",
            15: "#69fff8",
            16: "#f9ffa4",
            17: "#1c0dff",
        }
    )

    MODIS_LC_TYPE2 = (
        "MODIS/006/MCD12Q1",
        "LC_Type2",
        {
            0: "Water",
            1: "Evergreen conifer forest",
            2: "Evergreen broadleaf forest",
            3: "Deciduous conifer forest",
            4: "Deciduous broadleaf forest",
            5: "Mixed forest",
            6: "Closed shrubland",
            7: "Open shrubland",
            8: "Woody savanna",
            9: "Savanna",
            10: "Grassland",
            11: "Permanent wetland",
            12: "Cropland",
            13: "Urban",
            14: "Cropland and natural vegetation",
            15: "Barren",
        },
        {
            0: "#1c0dff",
            1: "#05450a",
            2: "#086a10",
            3: "#54a708",
            4: "#78d203",
            5: "#009900",
            6: "#c6b044",
            7: "#dcd159",
            8: "#dade48",
            9: "#fbff13",
            10: "#b6ff05",
            11: "#27ff87",
            12: "#c24f44",
            13: "#a5a5a5",
            14: "#ff6d4c",
            15: "#f9ffa4",
        }
    )

    MODIS_LC_TYPE3 = (
        "MODIS/006/MCD12Q1",
        "LC_Type3",
        {
            0: "Water",
            1: "Grassland",
            2: "Shrubland",
            3: "Crops",
            4: "Savannas",
            5: "Evergreen broadleaf",
            6: "Deciduous broadleaf",
            7: "Evergreen conifer",
            8: "Deciduous conifer",
            9: "Barren",
            10: "Urban"
        },
        {
            0: "#1c0dff",
            1: "#b6ff05",
            2: "#dcd159",
            3: "#c24f44",
            4: "#fbff13",
            5: "#086a10",
            6: "#78d203",
            7: "#05450a",
            8: "#54a708",
            9: "#f9ffa4",
            10: "#a5a5a5"
        }
    )

    CGLS_LC100 = (
        "COPERNICUS/Landcover/100m/Proba-V-C3/Global",
        "discrete_classification",
        {
            0: "Unknown",
            20: "Shrubs",
            30: "Herbaceous vegetation",
            40: "Cultivated",
            50: "Urban",
            60: "Bare",
            70: "Snow and ice",
            80: "Water body",
            90: "Herbaceous wetland",
            100: "Moss and lichen",
            111: "Closed forest, evergreen conifer",
            112: "Closed forest, evergreen broad leaf",
            113: "Closed forest, deciduous conifer",
            114: "Closed forest, deciduous broad leaf",
            115: "Closd forest, mixed",
            116: "Closed forest, other",
            121: "Open forest, evergreen conifer",
            122: "Open forest, evergreen broad leaf",
            123: "Open forest, deciduous conifer",
            124: "Open forest, deciduous broad leaf",
            125: "Open forest, mixed",
            126: "Open forest, other",
            200: "Ocean"
        },
        {
            0: "#282828",
            20: "#FFBB22",
            30: "#FFFF4C",
            40: "#F096FF",
            50: "#FA0000",
            60: "#B4B4B4",
            70: "#F0F0F0",
            80: "#0032C8",
            90: "#0096A0",
            100: "#FAE6A0",
            111: "#58481F",
            112: "#009900",
            113: "#70663E",
            114: "#00CC00",
            115: "#4E751F",
            116: "#007800",
            121: "#666000",
            122: "#8DB400",
            123: "#8D7400",
            124: "#A0DC00",
            125: "#929900",
            126: "#648C00",
            200: "#000080"
        }
    )

    @classmethod
    def names(cls):
        """
        Return string names of all datasets.
        """
        return [e.name for e in cls]

    @classmethod
    def get(cls, i=None):
        """
        Return object at a given index i or return all if i is none. 
        """
        if i is not None:
            return list(cls)[i]

        return [e for e in cls]
